/************************************************************
* 5-channel RGB DMX fixture controller                      *
*   DMX Controller with RGB+master intensity control plus   *
*   built-in programs.  EEPROM storage of DMX channel and   *
*   operating state.  Button + potentiometer controls and   *
*   4-digit 7-segment LED display.                          *
*                                                           *
* Developed by Dan Julio (www.danjuliodesigns.com)          *
* Rev 1.0 Dec 22, 2010                                      *
* Rev 1.1 Jan 13, 2011 - fixed bug with ScaleColor, other   *
*                        clean up, slightly tighten random  *
*                                                           *
* DMX receiver based on DMX-512 Reception, Developed by Max *
* Pierson                                                   *
*   Version Rev15 9 Oct 2010                                *
*                                                           *
* This software released under the same license as Max's:   *
*   "Released under the WTFPL license, although I would     *
*   appreciate Attribution and Share-Alike                  *
*   (blog.wingedvictorydesign.com)"                         *
*    www.danjuliodesigns.com                                *
*                                                           *
* Modified to decouple processing of data with ISR handling *
* of DMX frames.  The ISRs now run by themselves to detect  *
* each frame.  Remember to replace the HardwareSerial.cpp   *
* file in the Arduino release with the one Max provides.    *
*                                                           *
* This version compiled under Arduino 0016 although it      *
* should compile under later versions.                      *
*                                                           *
* User selectable modes (using push button)                 *
*  DMX: DMX control mode                                    *
*  DMX Addr Set: Set starting DMX address                   *
*    Short button press advances to next digit              *
*    Long button press exits DMX Addr Set mode to DMX mode  *
*    Pot sets value for blinking digit                      *
*  P1: Manual set color (pot sets color)                    *
*  P2: Strobe (using manually set P1 color, pot sets speed) *
*  P3: Sequential fade (pot sets speed)                     *
*  P4: Random fade (pot sets speed)                         *
*                                                           *
* Pressing the button quickly advances between DMX and P1-  *
* P4.  Pressing the button longer than two seconds in DMX   *
* mode causes the fixture to enter DMX Addr Set mode.       *           
*                                                           *
* DMX Channels in DMX mode                                  *
*  0: Mode                                                  *
*       0 -  63 : DMX Color control                         *
*      64 - 127 : Strobe using DMX Color control (value     *
*                 sets flashes/sec)                         *
*     128 - 191 : Sequential fade (value sets speed)        *
*     192 - 255 : Random fade (value sets speed)            *
*  1: Master Intensity                                      *
*  2: Red Channel                                           *
*  3: Green Channel                                         *
*  4: Blue Channel                                          *
*                                                           *
* EEPROM Storage                                            *
*   0: Magic Byte (set to 0xA5 when EEPROM initialized)     *
*   1: DMX Address High [8]                                 *
*   2: DMX Addres Low [7:0]                                 *
*   3: User-set Mode (values 0, 2-5 for UIMode)             *
*   4: P1 Color High [9:8] (used if stored mode is strobe)  *
*   5: P1 Color Low [7:0]                                   *
*                                                           *
* Arduino Port Connections:                                 *
*  D0: RX - TTL RX input from SN75176 driver                *
*  D1: TX - TTL TX output to SN75176 driver (unused)        *
*  D2: nRE - Active low Receiver Enable to SN75176          *
*  D3: DE - Active high Driver Enabled to SN75176           *
*  D4: LED Display 'a' cathode                              *
*  D5: Red PWM Output                                       *
*  D6: Green PWM Output                                     *
*  D7: LED Display 'b' cathode                              *
*  D8: LED Display 'c' cathode                              *
*  D9: Blue PWM Output                                      *
* D10: LED Display 'd' cathode                              *
* D11: LED Display 'e' cathode                              *
* D12: LED Display 'f' cathode                              *
* D13: LED Display 'g' cathode                              *
* A0(D14): LED Display digit 0 anode                        *
* A1(D15): LED Display digit 1 anode                        *
* A2(D16): LED Display digit 2 anode                        *
* A3(D17): LED Display digit 3 anode                        *
* A4(D18): Active low button input                          *
* A5: Potentiometer input (0 - VCC)                         *
*                                                           *
* Notes:                                                    *
*   1. DMX Addressing is limited to a maximum of 506 to     *
*      allow all 5 channels to inhabit legal addresses.     *
*      In practice the DMX receive code timing gets wonnky  *
*      at high addresses and DMX reception may not work.    *
*   2. TX capability is provided in hardware but unused by  *
*       this code (provides for future RDM capability).     *
*                                                           *
*************************************************************/
#include <EEPROM.h>


/******************************* Constant declarations *****************************/

#define NUMBER_OF_CHANNELS 5
#define MAX_DMX_ADDR 506

#define RECEIVER_OUTPUT_ENABLE 2
/* receiver output enable (pin2) on the SN75176.  
*  will be left low to set the SN75176 to receive data. */

#define DRIVER_OUTPUT_ENABLE 3
/* driver output enable (pin3) on the SN75176.  
*  will left low to disable driver output. */

#define RX_PIN 0   // serial receive pin, which takes the incoming data from the .
#define TX_PIN 1   // serial transmission pin

#define PWM_RED 5
#define PWM_GREEN 6
#define PWM_BLUE 9

#define LED_SEG_A 4
#define LED_SEG_B 7
#define LED_SEG_C 8
#define LED_SEG_D 10
#define LED_SEG_E 11
#define LED_SEG_F 12
#define LED_SEG_G 13
#define LED_DIG_0 14  // Repurposing A0-A3 as digital outputs
#define LED_DIG_1 15
#define LED_DIG_2 16
#define LED_DIG_3 17

#define MODE_BUTTON 18     // Repurposing A4 as digital input
#define AN_POT 5           // Analog input

#define DEBOUNCE_CNT 5     // number of evaluation intervals in a debounce period
#define BUTTON_LONG 80     // Long button press threshold in debounce periods

#define DIGIT_BLINK_PRD 12 // Number of complete multiplex periods in a full blink period

#define UI_MODE_DMX 0
#define UI_MODE_SET_ADDR 1
#define UI_MODE_PROG_1 2
#define UI_MODE_PROG_2 3
#define UI_MODE_PROG_3 4
#define UI_MODE_PROG_4 5

#define OP_MODE_DMX 0
#define OP_MODE_SET_ADDR 1
#define OP_MODE_PROG_1 2
#define OP_MODE_PROG_2 3
#define OP_MODE_PROG_3 4
#define OP_MODE_PROG_4 5

#define EEP_MAGIC_ADDR 0
#define EEP_DMX_H_ADDR 1
#define EEP_DMX_L_ADDR 2
#define EEP_USER_MODE  3
#define EEP_C_H_ADDR   4
#define EEP_C_L_ADDR   5
#define EEP_MAGIC_BYTE 0xA5 // The programmer leaves 0xFF as the default value

#define MODE_CHNG_CNT 200   // Number of evaluation cycles in one mode before updating EEPROM


/******************************* DMX ISR variable declarations ********************************/

volatile byte i = 0;                         //dummy variable for dmxvalue[]
volatile byte dmxreceived = 0;               //the latest received value
volatile unsigned int dmxcurrent = 0;        //counter variable that is incremented every time we receive a value.
volatile byte dmxvalue[NUMBER_OF_CHANNELS];  //stores the DMX values we're interested in using.
volatile boolean dmxnewvalue = false;            //set to 1 when new dmx values are received.

volatile byte zerocounter = 0;          
/* a counter to hold the number of zeros received in sequence on the serial receive pin.  
*  When we've received a minimum of 11 zeros in a row, we must be in a break.  As written,
*  the timer2 ISR actually checks for 20 zeros in a row, to ensure detecting an 88uS break. */

/*******************************  variable declarations **********************************/

unsigned int dmxaddress;       // The starting dmx address we will be listening to

byte CurRed;
byte CurGreen;
byte CurBlue;

byte NextRed;
byte NextGreen;
byte NextBlue;

unsigned long DeltaVal[3];     // Absolute delta between current and next color
unsigned long IncVal[3];       // Format {integer part[7:0] "." Fraction[15:0]}; range is 0x000000 - 0x010000 (0.0 - 1.0)
unsigned long InterpColor[3];  // Format {0x00, 0x00, Color[7:0] "." Fraction[15:0]}
boolean InterpSign[3];         // 1 = negative, 0 = positive
byte Major;                    // Major axis pointer: 0 = Red, 1 = Green, 2 = Blue

unsigned long PrevTime;
boolean Tick_5ms;

byte DebounceTimer=DEBOUNCE_CNT;
byte ButtonTimer=0;
boolean ButtonPrev = 0;
boolean ButtonReleased = 0;
boolean ButtonDown = 0;
boolean ButtonPressShort = 0;
boolean ButtonPressLong = 0;
unsigned int PotAvgBuff[4];
byte AvgBuffPtr = 0;

byte CurDigit = 0;    // Current digit to update (0-3)
byte DigitVals[4];
byte BlinkDigit;      // 0 = nothing blinking; 1-4 = blink digit 0-3
byte BlinkCount;

byte OPMode;
byte UIMode;
byte PrevUIMode;      // Used to detect mode changes to update EEPROM
byte ModeChangeTimer; // Used to time mode changes so we don't write to EEPROM unnecessarily

byte SetAddrDigit = 0;
byte Addr1s;
byte Addr10s;
byte Addr100s;
byte PrevPotDigit;   // 0 - 9 from the potentiometer setting

unsigned int PrevPotVal;  // Indexed color value for P1

byte ProgDelay;    // Number of evaluation intervals between program actions
unsigned int ProgIndex;


void setup() {
 CheckEEPsanity();
 InitController();
 InitVariables();
}


void loop()  { 
  if (dmxnewvalue) {
    EvalDMX();
    dmxnewvalue = false;
  }
  
  EvalTimeBase();
  if (Tick_5ms) {
    EvalControls();
    EvalDisplay();
    EvalLogic();
  }
}


void EvalTimeBase() {
  unsigned long CurTime = micros();
  unsigned long deltaT;
  
  if (CurTime < PrevTime) {
    // Clock rollover
    deltaT = PrevTime - CurTime;
  } else {
    deltaT = CurTime - PrevTime;
  }
  
  if (deltaT >= 5000) {
    Tick_5ms = true;
    PrevTime = CurTime;
  } else {
    Tick_5ms = false;
  }
}



// Timer2 compare match interrupt vector handler
ISR(TIMER2_COMPA_vect) {
  if (bitRead(PIND, PIND0)) {      // if a one is detected, we're not in a break, reset zerocounter.
    zerocounter = 0;
    }
  else {
    zerocounter++;                 // increment zerocounter if a zero is received.
    if (zerocounter == 20)         // if 20 0's are received in a row (80uS break)
      {   
      bitClear(TIMSK2, OCIE2A);    // disable this interrupt and enable reception interrupt now that we're in a break.
      bitSet(UCSR0B, RXCIE0);
      }
  }
} // end Timer2 ISR



ISR(USART_RX_vect){
  dmxreceived = UDR0;
  /* The receive buffer (UDR0) must be read during the reception ISR, or the ISR will just 
   *  execute again immediately upon exiting. */
 
  // Look for our data window
  if(++dmxcurrent > dmxaddress) {      // check if the current address is the one we want.
    dmxvalue[i++] = dmxreceived;
    if(i == NUMBER_OF_CHANNELS) {
      dmxnewvalue = true;              // set newvalue, so that the action loop can be executed.
      dmxcurrent = 0;                  // reset for the next frame
      zerocounter = 0;
      i = 0;
      bitClear(UCSR0B, RXCIE0);        // Disable receiving serial data
      bitSet(TIMSK2, OCIE2A);          // Enable Timer/Counter2 Output Compare Match A Interrupt for next break
    } 
  }
} // end ISR
